/*
 * Copyright (c) 2018-2021. data2viz sàrl.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */

package io.data2viz.scale

import io.data2viz.color.Colors
import io.data2viz.math.deg
import io.data2viz.math.pct
import io.data2viz.test.TestBase
import io.data2viz.test.shouldThrow
import kotlin.test.Test
import kotlin.test.assertTrue

class ScaleLinearTests : TestBase() {

    @Test
    fun copy(){
        val scale = Scales.Continuous.linear {
            domain = listOf(10.0, 1.0)
            range = listOf(.0, 100.0)
        }

        val copy = scale.copy()
        copy.domain shouldBe listOf(10.0, 1.0)
        copy.range shouldBe listOf(.0, 100.0)
    }


    @Test
    fun linear_no_interpolation() {
        val scale = Scales.Continuous.linear {
                    domain = listOf(.0, 1.0)
                    range = listOf(.0, 1.0)
                }

        scale(.5)       shouldBeClose .5
        scale(0.1)      shouldBeClose 0.1
        scale(0.133)    shouldBeClose 0.133
        scale(0.9)      shouldBeClose 0.9

        scale.domain = listOf(.0, 1.0)
        scale.range = listOf(1.0, .0)
        scale(.5)       shouldBeClose 0.5
        scale(0.1)      shouldBeClose 0.9
        scale(0.133)    shouldBeClose 0.867
        scale(0.9)      shouldBeClose  0.1          // TODO find why it is in fact 0.09999999999999998
    }

    @Test
    fun linear_Number() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, 100.0)
        scale.range = listOf(100.0, .0)
        scale(50.0) shouldBeClose 50.0
        scale(10.0) shouldBeClose 90.0
        scale(13.3) shouldBeClose 86.7
        scale(90.0) shouldBeClose 10.0
    }

    @Test
    fun linear_accept_Integer() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, 100.0)
        scale.range = listOf(100.0, .0)
        scale(50) shouldBeClose 50.0
        scale(10) shouldBeClose 90.0
        scale(13) shouldBeClose 87.0
        scale(90) shouldBeClose 10.0
    }


    @Test
    fun linear_Number_clamp_noclamp() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, 100.0)
        scale.range = listOf(.0, -100.0)
        scale(50.0) shouldBeClose -50.0
        scale(10.0) shouldBeClose -10.0
        scale(13.3) shouldBeClose -13.3
        scale(90.0) shouldBeClose -90.0
        scale(-20.0) shouldBeClose 20.0
        scale(132.0) shouldBeClose -132.0

        scale.clamp = true
        scale(50.0) shouldBeClose -50.0
        scale(10.0) shouldBeClose -10.0
        scale(13.3) shouldBeClose -13.3
        scale(90.0) shouldBeClose -90.0
        scale(-20.0) shouldBeClose 0.0
        scale(132.0) shouldBeClose -100.0
    }

    @Test
    fun linear_Number_invert() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, 100.0)
        scale.range = listOf(100.0, .0)
        scale.invert(50.0) shouldBeClose 50.0
        scale.invert(90.0) shouldBeClose 10.0
        scale.invert(86.7) shouldBeClose 13.3
        scale.invert(10.0) shouldBeClose 90.0
    }

    @Test
    fun linear_Number_bimap_ascending_descending_domain_and_range_invert() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, 100.0)
        scale.range = listOf(100.0, .0)
        scale.invert(50.0) shouldBeClose 50.0
        scale.invert(90.0) shouldBeClose 10.0
        scale.invert(86.7) shouldBeClose 13.3
        scale.invert(10.0) shouldBeClose 90.0

        scale.domain = listOf(100.0, .0)
        scale.range = listOf(.0, 100.0)
        scale.invert(50.0) shouldBeClose 50.0
        scale.invert(90.0) shouldBeClose 10.0
        scale.invert(86.7) shouldBeClose 13.3
        scale.invert(10.0) shouldBeClose 90.0

        scale.domain = listOf(100.0, .0)
        scale.range = listOf(100.0, .0)
        scale.invert(50.0) shouldBeClose 50.0
        scale.invert(90.0) shouldBeClose 90.0
        scale.invert(86.7) shouldBeClose 86.7
        scale.invert(10.0) shouldBeClose 10.0
    }

    @Test
    fun linear_Number_multiple_ranges() {
        val scale = Scales.Continuous.linear()

        // ASCENDING DOMAIN AND RANGE
        scale.domain = listOf(.0, 10.0, 20.0)
        scale.range = listOf(.0, 100.0, 300.0)
        scale(.0) shouldBeClose .0
        scale(10.0) shouldBeClose 100.0
        scale(20.0) shouldBeClose 300.0
        scale(5.0) shouldBeClose 50.0
        scale(15.0) shouldBeClose 200.0
        scale(17.0) shouldBeClose 240.0

        // DESCENDING DOMAIN
        // ASCENDING RANGE
        scale.domain = listOf(20.0, 10.0, .0)
        scale.range = listOf(.0, 100.0, 300.0)
        scale(.0) shouldBeClose 300.0
        scale(10.0) shouldBeClose 100.0
        scale(20.0) shouldBeClose .0
        scale(5.0) shouldBeClose 200.0
        scale(15.0) shouldBeClose 50.0
        scale(17.0) shouldBeClose 30.0

        // ASCENDING DOMAIN
        // DESCENDING RANGE
        scale.domain = listOf(.0, 10.0, 20.0)
        scale.range = listOf(300.0, 100.0, .0)
        scale(.0) shouldBeClose 300.0
        scale(10.0) shouldBeClose 100.0
        scale(20.0) shouldBeClose .0
        scale(5.0) shouldBeClose 200.0
        scale(15.0) shouldBeClose 50.0
        scale(17.0) shouldBeClose 30.0

        // DESCENDING DOMAIN
        // DESCENDING RANGE
        scale.domain = listOf(20.0, 10.0, .0)
        scale.range = listOf(300.0, 100.0, .0)
        scale(.0) shouldBeClose .0
        scale(10.0) shouldBeClose 100.0
        scale(20.0) shouldBeClose 300.0
        scale(5.0) shouldBeClose 50.0
        scale(15.0) shouldBeClose 200.0
        scale(17.0) shouldBeClose 240.0
    }

    @Test
    fun linear_Number_multiple_ranges_invert() {
        val scale = Scales.Continuous.linear()

        // ASCENDING DOMAIN AND RANGE
        scale.domain = listOf(.0, 10.0, 20.0)
        scale.range = listOf(.0, 100.0, 300.0)
        scale.invert(0.0) shouldBeClose 0.0
        scale.invert(100.0) shouldBeClose 10.0
        scale.invert(300.0) shouldBeClose 20.0
        scale.invert(50.0) shouldBeClose 5.0
        scale.invert(200.0) shouldBeClose 15.0
        scale.invert(240.0) shouldBeClose 17.0

        // DESCENDING DOMAIN
        // ASCENDING RANGE
        scale.domain = listOf(20.0, 10.0, .0)
        scale.range = listOf(300.0, 100.0, .0)
        scale.invert(0.0) shouldBeClose 0.0
        scale.invert(100.0) shouldBeClose 10.0
        scale.invert(300.0) shouldBeClose 20.0
        scale.invert(50.0) shouldBeClose 5.0
        scale.invert(200.0) shouldBeClose 15.0
        scale.invert(240.0) shouldBeClose 17.0

        // ASCENDING DOMAIN
        // DESCENDING RANGE
        scale.domain = listOf(.0, 10.0, 20.0)
        scale.range = listOf(300.0, 100.0, .0)
        scale.invert(.0) shouldBeClose 20.0
        scale.invert(100.0) shouldBeClose 10.0
        scale.invert(300.0) shouldBeClose .0
        scale.invert(50.0) shouldBeClose 15.0
        scale.invert(200.0) shouldBeClose 5.0
        scale.invert(240.0) shouldBeClose 3.0

        // DESCENDING DOMAIN
        // DESCENDING RANGE
        scale.domain = listOf(20.0, 10.0, .0)
        scale.range = listOf(300.0, 100.0, .0)
        scale.invert(.0) shouldBeClose .0
        scale.invert(100.0) shouldBeClose 10.0
        scale.invert(300.0) shouldBeClose 20.0
        scale.invert(50.0) shouldBeClose 5.0
        scale.invert(200.0) shouldBeClose 15.0
        scale.invert(240.0) shouldBeClose 17.0
    }

    @Test
    fun linear_Number_many_ranges_invert() {
        val scale = Scales.Continuous.linear()

        // ASCENDING DOMAIN AND DESCENDING RANGE
        scale.domain = listOf(.0,    10.0,   20.0,   30.0,   40.0,   50.0)
        scale.range = listOf(610.0,   310.0, 300.0,  110.0,  100.0,  -200.0)
        scale(0.0)  shouldBeClose 610.0
        scale(10.0) shouldBeClose 310.0
        scale(20.0) shouldBeClose 300.0
        scale(30.0) shouldBeClose 110.0
        scale(40.0) shouldBeClose 100.0
        scale(50.0) shouldBeClose -200.0
        scale.invert(610.0) shouldBeClose .0
        scale.invert(310.0) shouldBeClose 10.0
        scale.invert(300.0) shouldBeClose 20.0
        scale.invert(110.0) shouldBeClose 30.0
        scale.invert(100.0) shouldBeClose 40.0
        scale.invert(-200.0) shouldBeClose 50.0
        scale(-10.0)  shouldBeClose 910.0
        scale(5.0)  shouldBeClose 460.0
        scale(15.0) shouldBeClose 305.0
        scale(25.0) shouldBeClose 205.0
        scale(35.0) shouldBeClose 105.0
        scale(45.0) shouldBeClose -50.0
        scale(60.0) shouldBeClose -500.0
        scale.invert(910.0) shouldBeClose -10.0
        scale.invert(460.0) shouldBeClose 5.0
        scale.invert(305.0) shouldBeClose 15.0
        scale.invert(205.0) shouldBeClose 25.0
        scale.invert(105.0) shouldBeClose 35.0
        scale.invert(-50.0) shouldBeClose 45.0
        scale.invert(-500.0) shouldBeClose 60.0

        scale.clamp = true
        scale(-10.0)  shouldBeClose 610.0
        scale(60.0)  shouldBeClose -200.0
        scale.invert(900.0)  shouldBeClose .0
        scale.invert(-900.0)  shouldBeClose 50.0
    }

    @Test
    fun linear_double_rounded() {
        val scale = Scales.Continuous.linearRound()

        scale.domain = listOf(.0, 100.0)
        scale.range = listOf(.0, 100.0)
        scale(53.129) shouldBeClose 53.0
        scale(20.0255) shouldBeClose 20.0
        scale(93.899) shouldBeClose 94.0
    }

    @Test
    fun linear_HSL() {
        val scale = ScalesChromatic.Continuous.linearHSL()

        scale.domain = listOf(.0, 100.0)
        scale.range = listOf(Colors.hsl(0.deg, 100.pct, 100.pct), Colors.hsl(180.deg, 100.pct, 100.pct))
        scale(50.0) shouldBe Colors.hsl(90.deg, 100.pct, 100.pct)
        scale(20.0) shouldBe Colors.hsl(36.deg, 100.pct, 100.pct)
        scale(90.0) shouldBe Colors.hsl(162.deg, 100.pct, 100.pct)
    }

    /////////////// TESTS /////////////////////////////////////////

    @Test
    fun linear_expected_defaults_value() {
        val scale = Scales.Continuous.linear()

        scale.domain.size shouldBe 2
        scale.domain.first() shouldBeClose .0
        scale.domain.last() shouldBeClose 1.0
        scale.clamp shouldBe false
    }

    @Test
    fun linear_expected_scaling_with_defaults_value() {
        val scale = Scales.Continuous.linear()

        scale.range = listOf(1.0, 2.0)
        scale(.5) shouldBeClose 1.5
    }

    @Test
    fun linear_ignores_extra_range_values_for_smaller_domain() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(-10.0, .0)
        scale.range = listOf(0.0, 10.0, 2000.0)
        scale.clamp = true
        /*scale(-10.0) shouldBeClose .0
        scale(.0) shouldBeClose 10.0
        scale(-5.0) shouldBeClose 5.0
        scale(-15.0) shouldBeClose .0
        scale(10.0) shouldBeClose 10.0*/

        // changed behavior
        shouldThrow<IllegalStateException> { scale(.0) }
    }

    @Test
    fun linear_ignores_extra_domain_values_for_smaller_range() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(-10.0, .0, 2000.0)
        scale.range = listOf(0.0, 10.0)
        scale.clamp = true
        /*scale(-10.0) shouldBeClose .0
        scale(.0) shouldBeClose 10.0
        scale(-5.0) shouldBeClose 5.0
        scale(10.0) shouldBeClose 10.0*/

        // changed behavior
        shouldThrow<IllegalStateException> { scale(.0) }
    }

    @Test
    fun linear_maps_empty_domain_to_range_start_if_not_clamped() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, .0)
        scale.range = listOf(.0, 10.0)
        scale(1.0) shouldBeClose .0
        scale(-20.0) shouldBeClose .0
        scale(2000.0) shouldBeClose .0

        scale.domain = listOf(.0, .0)
        scale.range = listOf(10.0, .0)
        scale(1.0) shouldBeClose 10.0
        scale(-20.0) shouldBeClose 10.0
        scale(2000.0) shouldBeClose 10.0
    }

    @Test
    fun linear_maps_bilinear_domain_to_corresponding_range() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(1.0, 2.0)
        scale.range = listOf(.0, 1.0)
        scale(.5) shouldBeClose -.5
        scale(1.0) shouldBeClose .0
        scale(1.5) shouldBeClose .5
        scale(2.0) shouldBeClose 1.0
        scale(2.5) shouldBeClose 1.5
        scale.invert(-.5) shouldBeClose .5
        scale.invert(.0) shouldBeClose 1.0
        scale.invert(.5) shouldBeClose 1.5
        scale.invert(1.0) shouldBeClose 2.0
        scale.invert(1.5) shouldBeClose 2.5

        scale.clamp = true
        scale(.5) shouldBeClose .0
        scale(1.0) shouldBeClose .0
        scale(1.5) shouldBeClose .5
        scale(2.0) shouldBeClose 1.0
        scale(2.5) shouldBeClose 1.0
        scale.invert(-.5) shouldBeClose 1.0
        scale.invert(.0) shouldBeClose 1.0
        scale.invert(.5) shouldBeClose 1.5
        scale.invert(1.0) shouldBeClose 2.0
        scale.invert(1.5) shouldBeClose 2.0
    }

    @Test
    fun linear_maps_polylinear_domain_with_more_than_2_values_to_corresponding_range() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(1.0, 2.0, 4.0)
        scale.range = listOf(4.0, 2.0, 1.0)
        scale(1.5) shouldBeClose 3.0
        scale.invert(3.0) shouldBeClose 1.5
        scale(3.0) shouldBeClose 1.5
        scale.invert(1.5) shouldBeClose 3.0

        scale.domain = listOf(4.0, 2.0, 1.0)
        scale.range = listOf(1.0, 2.0, 4.0)
        scale(1.5) shouldBeClose 3.0
        scale.invert(3.0) shouldBeClose 1.5
        scale(3.0) shouldBeClose 1.5
        scale.invert(1.5) shouldBeClose 3.0
    }

    @Test
    fun linear_ticks_nice_OK() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(12.0, 87.0)
        scale.nice(5)
        scale.domain.size shouldBe 2
        scale.domain.first() shouldBeClose .0
        scale.domain.last() shouldBeClose 100.0

        scale.domain = listOf(12.0, 87.0)
        scale.nice(10)
        scale.domain.size shouldBe 2
        scale.domain.first() shouldBeClose 10.0
        scale.domain.last() shouldBeClose 90.0

        scale.domain = listOf(12.0, 87.0)
        scale.nice(100)
        scale.domain.size shouldBe 2
        scale.domain.first() shouldBeClose 12.0
        scale.domain.last() shouldBeClose 87.0
    }

    @Test
    fun linear_invert_maps_empty_range_to_domain_start() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(1.0, 2.0)
        scale.range = listOf(.0, .0)
        scale.invert(1.0) shouldBeClose 1.0
        scale.invert(.0) shouldBeClose 1.0
        scale.invert(-1.0) shouldBeClose 1.0

        scale.domain = listOf(2.0, 1.0)
        scale.range = listOf(.0, .0)
        scale.invert(1.0) shouldBeClose 2.0
        scale.invert(.0) shouldBeClose 2.0
        scale.invert(-1.0) shouldBeClose 2.0
    }

    @Test
    fun linear_accept_array_of_values() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf()
        scale.domain.size shouldBe 0
        scale.domain shouldBe listOf()

        scale.domain = listOf(1.0, 2.0)
        scale.domain.size shouldBe 2
        scale.domain shouldBe listOf(1.0, 2.0)

        scale.domain = listOf(1.0, 2.0, 3.0)
        scale.domain.size shouldBe 3
        scale.domain shouldBe listOf(1.0, 2.0, 3.0)
    }

    @Test
    fun linear_no_binding_for_domain_setter() {
        val scale = Scales.Continuous.linear()

        val array:MutableList<Double> = mutableListOf(1.0, 2.0)
        scale.domain = array
        scale.domain.size shouldBe 2
        scale.domain shouldBe listOf(1.0, 2.0)

        array.add(3.0)
        array.size shouldBe 3
        scale.domain.size shouldBe 2
        scale.domain shouldBe listOf(1.0, 2.0)
    }

    @Test
    fun linear_no_binding_for_domain_getter() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(1.0, 2.0)
        val array:MutableList<Double> = scale.domain.toMutableList()
        scale.domain.size shouldBe 2
        scale.domain shouldBe listOf(1.0, 2.0)

        array.add(3.0)
        array.size shouldBe 3
        scale.domain.size shouldBe 2
        scale.domain shouldBe listOf(1.0, 2.0)
    }

    @Test
    fun linear_no_binding_for_range_setter() {
        val scale = Scales.Continuous.linear()

        val array:MutableList<Double> = mutableListOf(1.0, 2.0)
        scale.range = array
        scale.range.size shouldBe 2
        scale.range shouldBe listOf(1.0, 2.0)

        array.add(3.0)
        array.size shouldBe 3
        scale.range.size shouldBe 2
        scale.range shouldBe listOf(1.0, 2.0)
    }

    @Test
    fun linear_no_binding_for_range_getter() {
        val scale = Scales.Continuous.linear()

        scale.range = listOf(1.0, 2.0)
        val array:MutableList<Double> = scale.range.toMutableList()
        scale.range.size shouldBe 2
        scale.range shouldBe listOf(1.0, 2.0)

        array.add(3.0)
        array.size shouldBe 3
        scale.range.size shouldBe 2
        scale.range shouldBe listOf(1.0, 2.0)
    }

    // TODO tape("linear.rangeRound(range) is an alias for linear.range(range).interpolate(interpolateRound)", function(test) {

    @Test
    fun linear_clamp_is_false_by_default() {
        val scale = Scales.Continuous.linear()

        scale.clamp shouldBe false
        scale.range = listOf(10.0, 20.0)
        scale(2.0) shouldBeClose 30.0
        scale(-1.0) shouldBeClose .0
        scale.invert(30.0) shouldBeClose 2.0
        scale.invert(.0) shouldBeClose -1.0
    }

    @Test
    fun linear_clamp_restricts_output_to_the_range() {
        val scale = Scales.Continuous.linear()

        scale.clamp = true
        scale.range = listOf(10.0, 20.0)
        scale(2.0) shouldBeClose 20.0
        scale(-1.0) shouldBeClose 10.0
    }

    @Test
    fun linear_clamp_restricts_input_to_the_domain() {
        val scale = Scales.Continuous.linear()

        scale.clamp = true
        scale.range = listOf(10.0, 20.0)
        scale.invert(30.0) shouldBeClose 1.0
        scale.invert(.0) shouldBeClose .0
    }

    // TODO tape("linear.interpolate(interpolate) takes a custom interpolator factory", function(test) {

    @Test
    fun linear_nice_is_10_ticks_by_default() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, .96)
        scale.nice()
        scale.domain.first() shouldBeClose .0
        scale.domain.last() shouldBeClose 1.0

        scale.domain = listOf(.0, 96.0)
        scale.nice()
        scale.domain.first() shouldBeClose .0
        scale.domain.last() shouldBeClose 100.0
    }

    @Test
    fun linear_nice_extends_domain_to_match_desired_ticks() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, .96)
        scale.nice(10)
        scale.domain shouldBe listOf(.0, 1.0)

        scale.domain = listOf(.0, 96.0)
        scale.nice(10)
        scale.domain shouldBe listOf(.0, 100.0)

        scale.domain = listOf(.96, .0)
        scale.nice(10)
        scale.domain shouldBe listOf(1.0, .0)

        scale.domain = listOf(96.0, .0)
        scale.nice(10)
        scale.domain shouldBe listOf(100.0, .0)

        scale.domain = listOf(.0, -.96)
        scale.nice(10)
        scale.domain shouldBe listOf(.0, -1.0)

        scale.domain = listOf(.0, -96.0)
        scale.nice(10)
        scale.domain shouldBe listOf(.0, -100.0)

        scale.domain = listOf(-.96, .0)
        scale.nice(10)
        scale.domain shouldBe listOf(-1.0, .0)

        scale.domain = listOf(-96.0, .0)
        scale.nice(10)
        scale.domain shouldBe listOf(-100.0, .0)

        scale.domain = listOf(-.1, 51.1)
        scale.nice(8)
        scale.domain shouldBe listOf(-10.0, 60.0)
    }

    @Test
    fun linear_nice_extends_domain_to_round_numbers() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(1.1, 10.9)
        scale.nice(10)
        scale.domain shouldBe listOf(1.0, 11.0)

        scale.domain = listOf(10.9, 1.1)
        scale.nice(10)
        scale.domain shouldBe listOf(11.0, 1.0)

        scale.domain = listOf(.7, 11.001)
        scale.nice(10)
        scale.domain shouldBe listOf(.0, 12.0)

        scale.domain = listOf(123.1, 6.7)
        scale.nice(10)
        scale.domain shouldBe listOf(130.0, .0)

        scale.domain = listOf(.0, .49)
        scale.nice(10)
        scale.domain shouldBe listOf(.0, .5)

        scale.domain = listOf(.0, 14.1)
        scale.nice(5)
        scale.domain shouldBe listOf(.0, 20.0)

        scale.domain = listOf(.0, 15.0)
        scale.nice(5)
        scale.domain shouldBe listOf(.0, 20.0)
        scale.range = listOf(.0, 100.0)
        scale(20.0) shouldBeClose 100.0
    }

    @Test
    fun linear_nice_no_effect_on_invalid_domains() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, .0)
        scale.nice(10)
        scale.domain shouldBe listOf(.0, .0)

        scale.domain = listOf(.5, .5)
        scale.nice(10)
        scale.domain shouldBe listOf(.5, .5)
    }

    @Test
    fun linear_nice_only_the_extent_of_a_polydomain() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(1.1, 1.0, 2.0, 3.0, 10.9)
        scale.nice(10)
        scale.domain shouldBe listOf(1.0, 1.0, 2.0, 3.0, 11.0)

        scale.domain = listOf(123.1, 1.0, 2.0, 3.0, -0.9)
        scale.nice(10)
        scale.domain shouldBe listOf(130.0, 1.0, 2.0, 3.0, -10.0)
    }

    @Test
    fun linear_nice_accept_a_tick_to_control_nicing_step() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(12.0, 87.0)
        scale.nice(5)
        scale.domain shouldBe listOf(.0, 100.0)

        scale.domain = listOf(12.0, 87.0)
        scale.nice(10)
        scale.domain shouldBe listOf(10.0, 90.0)

        scale.domain = listOf(12.0, 87.0)
        scale.nice(100)
        scale.domain shouldBe listOf(12.0, 87.0)
    }

    @Test
    fun linear_ticks_return_expected_ticks() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, 1.0)
        scale.ticks(10).size shouldBe 11
        scale.ticks(10) shouldBe listOf(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
        scale.ticks(9) shouldBe  listOf(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
        scale.ticks(8) shouldBe  listOf(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
        scale.ticks(7) shouldBe  listOf(0.0,      0.2,      0.4,      0.6,      0.8,      1.0)
        scale.ticks(6) shouldBe  listOf(0.0,      0.2,      0.4,      0.6,      0.8,      1.0)
        scale.ticks(5) shouldBe  listOf(0.0,      0.2,      0.4,      0.6,      0.8,      1.0)
        scale.ticks(4) shouldBe  listOf(0.0,      0.2,      0.4,      0.6,      0.8,      1.0)
        scale.ticks(3) shouldBe  listOf(0.0,                     0.5,                     1.0)
        scale.ticks(2) shouldBe  listOf(0.0,                     0.5,                     1.0)
        scale.ticks(1) shouldBe  listOf(0.0,                                              1.0)

        scale.domain = listOf(-100.0, 100.0)
        scale.ticks(10) shouldBe listOf(-100.0, -80.0, -60.0,        -40.0, -20.0, .0, 20.0, 40.0,       60.0, 80.0, 100.0)
        scale.ticks(9) shouldBe  listOf(-100.0, -80.0, -60.0,        -40.0, -20.0, .0, 20.0, 40.0,       60.0, 80.0, 100.0)
        scale.ticks(8) shouldBe  listOf(-100.0, -80.0, -60.0,        -40.0, -20.0, .0, 20.0, 40.0,       60.0, 80.0, 100.0)
        scale.ticks(7) shouldBe  listOf(-100.0, -80.0, -60.0,        -40.0, -20.0, .0, 20.0, 40.0,       60.0, 80.0, 100.0)
        scale.ticks(6) shouldBe  listOf(-100.0,               -50.0,               .0,             50.0,             100.0)
        scale.ticks(5) shouldBe  listOf(-100.0,               -50.0,               .0,             50.0,             100.0)
        scale.ticks(4) shouldBe  listOf(-100.0,               -50.0,               .0,             50.0,             100.0)
        scale.ticks(3) shouldBe  listOf(-100.0,               -50.0,               .0,             50.0,             100.0)
        scale.ticks(2) shouldBe  listOf(-100.0,                                    .0,                               100.0)
        scale.ticks(1) shouldBe  listOf(                                           .0                                     )
    }

    @Test
    fun linear_ticks_descending_domain_return_expected_ticks() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(1.0, .0)
        scale.ticks(10).size shouldBe 11
        scale.ticks(10) shouldBe listOf(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0).reversed()
        scale.ticks(9) shouldBe  listOf(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0).reversed()
        scale.ticks(8) shouldBe  listOf(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0).reversed()
        scale.ticks(7) shouldBe  listOf(0.0,      0.2,      0.4,      0.6,      0.8,      1.0).reversed()
        scale.ticks(6) shouldBe  listOf(0.0,      0.2,      0.4,      0.6,      0.8,      1.0).reversed()
        scale.ticks(5) shouldBe  listOf(0.0,      0.2,      0.4,      0.6,      0.8,      1.0).reversed()
        scale.ticks(4) shouldBe  listOf(0.0,      0.2,      0.4,      0.6,      0.8,      1.0).reversed()
        scale.ticks(3) shouldBe  listOf(0.0,                     0.5,                     1.0).reversed()
        scale.ticks(2) shouldBe  listOf(0.0,                     0.5,                     1.0).reversed()
        scale.ticks(1) shouldBe  listOf(0.0,                                              1.0).reversed()

        scale.domain = listOf(100.0, -100.0)
        scale.ticks(10) shouldBe listOf(-100.0, -80.0, -60.0,        -40.0, -20.0, .0, 20.0, 40.0,       60.0, 80.0, 100.0).reversed()
        scale.ticks(9) shouldBe  listOf(-100.0, -80.0, -60.0,        -40.0, -20.0, .0, 20.0, 40.0,       60.0, 80.0, 100.0).reversed()
        scale.ticks(8) shouldBe  listOf(-100.0, -80.0, -60.0,        -40.0, -20.0, .0, 20.0, 40.0,       60.0, 80.0, 100.0).reversed()
        scale.ticks(7) shouldBe  listOf(-100.0, -80.0, -60.0,        -40.0, -20.0, .0, 20.0, 40.0,       60.0, 80.0, 100.0).reversed()
        scale.ticks(6) shouldBe  listOf(-100.0,               -50.0,               .0,             50.0,             100.0).reversed()
        scale.ticks(5) shouldBe  listOf(-100.0,               -50.0,               .0,             50.0,             100.0).reversed()
        scale.ticks(4) shouldBe  listOf(-100.0,               -50.0,               .0,             50.0,             100.0).reversed()
        scale.ticks(3) shouldBe  listOf(-100.0,               -50.0,               .0,             50.0,             100.0).reversed()
        scale.ticks(2) shouldBe  listOf(-100.0,                                    .0,                               100.0).reversed()
        scale.ticks(1) shouldBe  listOf(                                           .0                                     ).reversed()
    }

    @Test
    fun linear_ticks_polylinear_domain_return_expected_ticks() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, 0.25, 0.9, 1.0)
        scale.ticks(10).size shouldBe 11
        scale.ticks(10) shouldBe listOf(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
        scale.ticks(9) shouldBe  listOf(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
        scale.ticks(8) shouldBe  listOf(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
        scale.ticks(7) shouldBe  listOf(0.0,      0.2,      0.4,      0.6,      0.8,      1.0)
        scale.ticks(6) shouldBe  listOf(0.0,      0.2,      0.4,      0.6,      0.8,      1.0)
        scale.ticks(5) shouldBe  listOf(0.0,      0.2,      0.4,      0.6,      0.8,      1.0)
        scale.ticks(4) shouldBe  listOf(0.0,      0.2,      0.4,      0.6,      0.8,      1.0)
        scale.ticks(3) shouldBe  listOf(0.0,                     0.5,                     1.0)
        scale.ticks(2) shouldBe  listOf(0.0,                     0.5,                     1.0)
        scale.ticks(1) shouldBe  listOf(0.0,                                              1.0)

        scale.domain = listOf(-100.0, .0, 100.0)
        scale.ticks(10) shouldBe listOf(-100.0, -80.0, -60.0,        -40.0, -20.0, .0, 20.0, 40.0,       60.0, 80.0, 100.0)
        scale.ticks(9) shouldBe  listOf(-100.0, -80.0, -60.0,        -40.0, -20.0, .0, 20.0, 40.0,       60.0, 80.0, 100.0)
        scale.ticks(8) shouldBe  listOf(-100.0, -80.0, -60.0,        -40.0, -20.0, .0, 20.0, 40.0,       60.0, 80.0, 100.0)
        scale.ticks(7) shouldBe  listOf(-100.0, -80.0, -60.0,        -40.0, -20.0, .0, 20.0, 40.0,       60.0, 80.0, 100.0)
        scale.ticks(6) shouldBe  listOf(-100.0,               -50.0,               .0,             50.0,             100.0)
        scale.ticks(5) shouldBe  listOf(-100.0,               -50.0,               .0,             50.0,             100.0)
        scale.ticks(4) shouldBe  listOf(-100.0,               -50.0,               .0,             50.0,             100.0)
        scale.ticks(3) shouldBe  listOf(-100.0,               -50.0,               .0,             50.0,             100.0)
        scale.ticks(2) shouldBe  listOf(-100.0,                                    .0,                               100.0)
        scale.ticks(1) shouldBe  listOf(                                           .0                                     )
    }

    @Test
    fun linear_ticks_count_should_be_greater_than_zero() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, 1.0)
        scale.ticks(0).size shouldBe 0
        scale.ticks(0) shouldBe listOf<Double>()
        scale.ticks(-2).size shouldBe 0
        scale.ticks(-2) shouldBe listOf<Double>()

        // TODO shouldThrow ??
//        shouldThrow<IllegalArgumentException> { scale.ticks(-2) }
//        shouldThrow<IllegalArgumentException> { scale.ticks(Int.MIN_VALUE) }
    }

    @Test
    fun linear_ticks_count_without_arguments_should_be_count_equals_10() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(.0, 1.0)
        scale.ticks().size shouldBe scale.ticks(10).size
    }

    @Test
    fun tick_values_does_not_contains_values_outside_of_domain() {
        val scale = Scales.Continuous.linear()

        scale.domain = listOf(14.649324193073626, 17.002125621713716)
        scale.ticks(10) shouldBe listOf(14.8, 15.0, 15.2, 15.4, 15.6, 15.8, 16.0, 16.2, 16.4, 16.6, 16.8, 17.0)

        scale.domain = listOf(17.002125621713716, 14.649324193073626)
        scale.ticks(10) shouldBe listOf(14.8, 15.0, 15.2, 15.4, 15.6, 15.8, 16.0, 16.2, 16.4, 16.6, 16.8, 17.0).reversed()

        scale.domain = listOf(14.649324193073626, 17.2)
        scale.ticks(10) shouldBe listOf(14.8, 15.0, 15.2, 15.4, 15.6, 15.8, 16.0, 16.2, 16.4, 16.6, 16.8, 17.0, 17.2)

        scale.domain = listOf(14.6, 17.2)
        scale.ticks(10) shouldBe listOf(14.6, 14.8, 15.0, 15.2, 15.4, 15.6, 15.8, 16.0, 16.2, 16.4, 16.6, 16.8, 17.0, 17.2)

        scale.domain = listOf(14.6, 17.0002)
        scale.ticks(10) shouldBe listOf(14.6, 14.8, 15.0, 15.2, 15.4, 15.6, 15.8, 16.0, 16.2, 16.4, 16.6, 16.8, 17.0)

        scale.domain = listOf(1.0006002, 1.0007002)
        scale.ticks(10) shouldBe listOf(1.00061, 1.00062, 1.00063, 1.00064, 1.00065, 1.00066, 1.00067, 1.00068, 1.00069, 1.0007)

        scale.domain = listOf(1.0007002, 1.0006002)
        scale.ticks(10) shouldBe listOf(1.00061, 1.00062, 1.00063, 1.00064, 1.00065, 1.00066, 1.00067, 1.00068, 1.00069, 1.0007).reversed()
    }

    // TODO "linear.tickFormat(count) returns a format suitable for the ticks" etc...
    // TODO "linear.copy() returns a copy with changes to the interpolator are isolated"
}
